# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: wandb/proto/wandb_telemetry.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='wandb/proto/wandb_telemetry.proto',
  package='wandb_internal',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=b'\n!wandb/proto/wandb_telemetry.proto\x12\x0ewandb_internal\"\x87\x02\n\x0fTelemetryRecord\x12-\n\x0cimports_init\x18\x01 \x01(\x0b\x32\x17.wandb_internal.Imports\x12/\n\x0eimports_finish\x18\x02 \x01(\x0b\x32\x17.wandb_internal.Imports\x12(\n\x07\x66\x65\x61ture\x18\x03 \x01(\x0b\x32\x17.wandb_internal.Feature\x12\x16\n\x0epython_version\x18\x04 \x01(\t\x12\x13\n\x0b\x63li_version\x18\x05 \x01(\t\x12\x1b\n\x13huggingface_version\x18\x06 \x01(\t\x12 \n\x03\x65nv\x18\x08 \x01(\x0b\x32\x13.wandb_internal.Env\"\xe6\x01\n\x07Imports\x12\r\n\x05torch\x18\x01 \x01(\x08\x12\r\n\x05keras\x18\x02 \x01(\x08\x12\x12\n\ntensorflow\x18\x03 \x01(\x08\x12\x0e\n\x06\x66\x61stai\x18\x04 \x01(\x08\x12\x0f\n\x07sklearn\x18\x05 \x01(\x08\x12\x0f\n\x07xgboost\x18\x06 \x01(\x08\x12\x10\n\x08\x63\x61tboost\x18\x07 \x01(\x08\x12\x10\n\x08lightgbm\x18\x08 \x01(\x08\x12\x19\n\x11pytorch_lightning\x18\t \x01(\x08\x12\x16\n\x0epytorch_ignite\x18\n \x01(\x08\x12 \n\x18transformers_huggingface\x18\x0b \x01(\x08\"\xb6\x01\n\x07\x46\x65\x61ture\x12\r\n\x05watch\x18\x01 \x01(\x08\x12\x0e\n\x06\x66inish\x18\x02 \x01(\x08\x12\x0c\n\x04save\x18\x03 \x01(\x08\x12\x0f\n\x07offline\x18\x04 \x01(\x08\x12\x0f\n\x07resumed\x18\x05 \x01(\x08\x12\x0c\n\x04grpc\x18\x06 \x01(\x08\x12\x0e\n\x06metric\x18\x07 \x01(\x08\x12\r\n\x05keras\x18\x08 \x01(\x08\x12\x11\n\tsagemaker\x18\t \x01(\x08\x12\x1c\n\x14\x61rtifact_incremental\x18\n \x01(\x08\"\xa0\x01\n\x03\x45nv\x12\x0f\n\x07jupyter\x18\x01 \x01(\x08\x12\x0e\n\x06kaggle\x18\x02 \x01(\x08\x12\x0f\n\x07windows\x18\x03 \x01(\x08\x12\x0e\n\x06m1_gpu\x18\x04 \x01(\x08\x12\x13\n\x0bstart_spawn\x18\x05 \x01(\x08\x12\x12\n\nstart_fork\x18\x06 \x01(\x08\x12\x18\n\x10start_forkserver\x18\x07 \x01(\x08\x12\x14\n\x0cstart_thread\x18\x08 \x01(\x08\x62\x06proto3'
)




_TELEMETRYRECORD = _descriptor.Descriptor(
  name='TelemetryRecord',
  full_name='wandb_internal.TelemetryRecord',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='imports_init', full_name='wandb_internal.TelemetryRecord.imports_init', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='imports_finish', full_name='wandb_internal.TelemetryRecord.imports_finish', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='feature', full_name='wandb_internal.TelemetryRecord.feature', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='python_version', full_name='wandb_internal.TelemetryRecord.python_version', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='cli_version', full_name='wandb_internal.TelemetryRecord.cli_version', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='huggingface_version', full_name='wandb_internal.TelemetryRecord.huggingface_version', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='env', full_name='wandb_internal.TelemetryRecord.env', index=6,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=54,
  serialized_end=317,
)


_IMPORTS = _descriptor.Descriptor(
  name='Imports',
  full_name='wandb_internal.Imports',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='torch', full_name='wandb_internal.Imports.torch', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='keras', full_name='wandb_internal.Imports.keras', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tensorflow', full_name='wandb_internal.Imports.tensorflow', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='fastai', full_name='wandb_internal.Imports.fastai', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sklearn', full_name='wandb_internal.Imports.sklearn', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='xgboost', full_name='wandb_internal.Imports.xgboost', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='catboost', full_name='wandb_internal.Imports.catboost', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='lightgbm', full_name='wandb_internal.Imports.lightgbm', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pytorch_lightning', full_name='wandb_internal.Imports.pytorch_lightning', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pytorch_ignite', full_name='wandb_internal.Imports.pytorch_ignite', index=9,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='transformers_huggingface', full_name='wandb_internal.Imports.transformers_huggingface', index=10,
      number=11, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=320,
  serialized_end=550,
)


_FEATURE = _descriptor.Descriptor(
  name='Feature',
  full_name='wandb_internal.Feature',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='watch', full_name='wandb_internal.Feature.watch', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='finish', full_name='wandb_internal.Feature.finish', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='save', full_name='wandb_internal.Feature.save', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='offline', full_name='wandb_internal.Feature.offline', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='resumed', full_name='wandb_internal.Feature.resumed', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='grpc', full_name='wandb_internal.Feature.grpc', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='metric', full_name='wandb_internal.Feature.metric', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='keras', full_name='wandb_internal.Feature.keras', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sagemaker', full_name='wandb_internal.Feature.sagemaker', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='artifact_incremental', full_name='wandb_internal.Feature.artifact_incremental', index=9,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=553,
  serialized_end=735,
)


_ENV = _descriptor.Descriptor(
  name='Env',
  full_name='wandb_internal.Env',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='jupyter', full_name='wandb_internal.Env.jupyter', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kaggle', full_name='wandb_internal.Env.kaggle', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='windows', full_name='wandb_internal.Env.windows', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='m1_gpu', full_name='wandb_internal.Env.m1_gpu', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='start_spawn', full_name='wandb_internal.Env.start_spawn', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='start_fork', full_name='wandb_internal.Env.start_fork', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='start_forkserver', full_name='wandb_internal.Env.start_forkserver', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='start_thread', full_name='wandb_internal.Env.start_thread', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=738,
  serialized_end=898,
)

_TELEMETRYRECORD.fields_by_name['imports_init'].message_type = _IMPORTS
_TELEMETRYRECORD.fields_by_name['imports_finish'].message_type = _IMPORTS
_TELEMETRYRECORD.fields_by_name['feature'].message_type = _FEATURE
_TELEMETRYRECORD.fields_by_name['env'].message_type = _ENV
DESCRIPTOR.message_types_by_name['TelemetryRecord'] = _TELEMETRYRECORD
DESCRIPTOR.message_types_by_name['Imports'] = _IMPORTS
DESCRIPTOR.message_types_by_name['Feature'] = _FEATURE
DESCRIPTOR.message_types_by_name['Env'] = _ENV
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

TelemetryRecord = _reflection.GeneratedProtocolMessageType('TelemetryRecord', (_message.Message,), {
  'DESCRIPTOR' : _TELEMETRYRECORD,
  '__module__' : 'wandb.proto.wandb_telemetry_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.TelemetryRecord)
  })
_sym_db.RegisterMessage(TelemetryRecord)

Imports = _reflection.GeneratedProtocolMessageType('Imports', (_message.Message,), {
  'DESCRIPTOR' : _IMPORTS,
  '__module__' : 'wandb.proto.wandb_telemetry_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.Imports)
  })
_sym_db.RegisterMessage(Imports)

Feature = _reflection.GeneratedProtocolMessageType('Feature', (_message.Message,), {
  'DESCRIPTOR' : _FEATURE,
  '__module__' : 'wandb.proto.wandb_telemetry_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.Feature)
  })
_sym_db.RegisterMessage(Feature)

Env = _reflection.GeneratedProtocolMessageType('Env', (_message.Message,), {
  'DESCRIPTOR' : _ENV,
  '__module__' : 'wandb.proto.wandb_telemetry_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.Env)
  })
_sym_db.RegisterMessage(Env)


# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
"""Sweep config tune interface.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from wandb.sweeps.config.tune import tune

run = tune.run
loguniform = tune.loguniform
uniform = tune.uniform

__all__ = [
    "run",
    "uniform",
    "loguniform",
]




#
# -*- coding: utf-8 -*-
"""
History tracks logged data over time. To use history from your script, call
wandb.log({"key": value}) at a single time step or multiple times in your
training loop. This generates a time series of saved scalars or media that is
saved to history.

In the UI, if you log a scalar at multiple timesteps W&B will render these
history metrics as line plots by default. If you log a single value in history,
compare across runs with a bar chart.

It's often useful to track a full time series as well as a single summary value.
For example, accuracy at every step in History and best accuracy in Summary.
By default, Summary is set to the final value of History.
"""

import time

from wandb.wandb_torch import TorchHistory


class History(object):
    """Time series data for Runs. This is essentially a list of dicts where each
        dict is a set of summary statistics logged.
    """

    def __init__(self, run):
        self._run = run
        self._step = 0
        self._data = dict()
        self._callback = None
        self._torch = None
        self.compute = True

    def _set_callback(self, cb):
        self._callback = cb

    def _row_update(self, row):
        self._data.update(row)

    def _row_add(self, row):
        self._data.update(row)
        self._flush()
        self._step += 1

    def _update_step(self):
        """Called after receiving the run from the internal process"""
        self._step = self._run.starting_step

    def _flush(self):
        if len(self._data) > 0:
            self._data["_step"] = self._step
            self._data["_runtime"] = int(
                self._data.get("_runtime", time.time() - self.start_time)
            )
            self._data["_timestamp"] = int(self._data.get("_timestamp", time.time()))
            if self._callback:
                self._callback(row=self._data, step=self._step)
            self._data = dict()

    @property
    def start_time(self):
        return self._run.start_time

    def add(self, d):
        self._row_add(d)

    @property
    def torch(self):
        if self._torch is None:
            self._torch = TorchHistory(self)
        return self._torch



import numpy as np
from numpy.random import uniform
from numpy.random import randint
from wandb.sweeps import envelope_stopping


def synthetic_loss(start, asympt, decay, noise, length):
    val = start
    history = []
    metric = []
    for ii in range(length):
        history.append(val)
        val += uniform(-noise, noise)
        val -= (val - asympt) * decay
    return history


def synthetic_loss_random():
    return synthetic_loss(
        uniform(4, 20), uniform(1, 2), uniform(0.05, 0.4), 0.5, randint(10, 20)
    )


def synthetic_loss_family(num):
    histories = []
    for ii in range(num):
        history = synthetic_loss(
            uniform(4, 20), uniform(2, 3.), uniform(0.05, 0.4), 0.5, randint(10, 20)
        )
        histories.append(history)
    return histories


def test_envelope_terminate():
    hs = synthetic_loss_family(20)
    m = []
    for h in hs:
        m.append(min(h))
    top_hs = envelope_stopping.histories_for_top_n(hs, m, 5)
    envelope = envelope_stopping.envelope_from_histories(top_hs, 30)
    tries = 10
    for i in range(tries):
        new_history = synthetic_loss(
            20 + uniform(4, 20), 20., uniform(0.05, 0.4), 0.5, randint(10, 40)
        )
        assert (not envelope_stopping.is_inside_envelope(new_history, envelope))
    for i in range(tries):
        new_history = synthetic_loss(
            uniform(0, 2.), 20., uniform(0, 1.), 0.5, randint(10, 50)
        )
        print(new_history)
        print(envelope)
        assert envelope_stopping.is_inside_envelope(new_history, envelope)



"""Base SweepConfig classes.
"""

import wandb
import yaml

from six.moves import UserDict


class SweepConfig(UserDict):
    def __init__(self, d):
        super(SweepConfig, self).__init__(d)

    def __str__(self):
        return yaml.safe_dump(self.data)

    def save(self, filename):
        with open(filename, "w") as outfile:
            yaml.safe_dump(self.data, outfile, default_flow_style=False)

    def set_local(self):
        self.data.update(dict(controller=dict(type="local")))
        return self

    def set_name(self, name):
        self.data.update(dict(name=name))
        return self

    def set_settings(self, settings):
        self.data.update(dict(settings=settings))
        return self

    def set(self, **kwargs):
        local = kwargs.pop("local", None)
        name = kwargs.pop("name", None)
        settings = kwargs.pop("settings", None)
        if local:
            self.set_local()
        if name:
            self.set_name(name)
        if name:
            self.set_settings(settings)
        for k in kwargs.keys():
            wandb.termwarn("Unsupported parameter passed to SweepConfig set(): {}".format(k))
        return self


class SweepConfigElement:
    _version_dict = {}
    def __init__(self, module=None, version=None):
        self._module = module
        self._version = version
        self._version_dict.setdefault("wandb", wandb.__version__)
        if module and version:
            self._version_dict.setdefault(module, version)

    def _config(self, base, args, kwargs, root=False):
        kwargs = {k:v for k, v in kwargs.items() if v is not None and k != "self"}
        # remove kwargs if empty
        if kwargs.get("kwargs") == {}:
            del kwargs["kwargs"]
        # if only kwargs specified and only two keys "args" and "kargs"
        special = not args and set(kwargs.keys()) == set(("args", "kwargs"))
        if args and kwargs or special:
            d = dict(args=args, kwargs=kwargs)
        elif args:
            d = args
        else:
            d = kwargs
        if base:
            if self._module:
                base = self._module + "." + base
            d = {base: d}
        if root:
            # HACK(jhr): move tune.run to tune
            d = d["tune.run"]
            d = dict(tune=d)
            for m, v in self._version_dict.items():
                d["tune"].setdefault("_wandb", {})
                d["tune"]["_wandb"].setdefault("versions", {})
                d["tune"]["_wandb"]["versions"][m] = v
            return SweepConfig(d)
        return d



# -*- coding: utf-8 -*-
"""
`expand` type question
"""
from __future__ import print_function, unicode_literals

import sys

from prompt_toolkit.application import Application
from prompt_toolkit.key_binding.manager import KeyBindingManager
from prompt_toolkit.keys import Keys
from prompt_toolkit.layout.containers import Window
from prompt_toolkit.filters import IsDone
from prompt_toolkit.layout.controls import TokenListControl
from prompt_toolkit.layout.containers import ConditionalContainer, HSplit
from prompt_toolkit.layout.dimension import LayoutDimension as D
from prompt_toolkit.token import Token

from .. import PromptParameterException
from ..separator import Separator
from .common import default_style
from .common import if_mousedown


PY3 = sys.version_info[0] >= 3

if PY3:
    basestring = str


# custom control based on TokenListControl

class InquirerControl(TokenListControl):
    def __init__(self, choices, default=None, **kwargs):
        self.pointer_index = 0
        self.answered = False
        self._init_choices(choices, default)
        self._help_active = False  # help is activated via 'h' key
        super(InquirerControl, self).__init__(self._get_choice_tokens,
                                              **kwargs)

    def _init_choices(self, choices, default=None):
        # helper to convert from question format to internal format

        self.choices = []  # list (key, name, value)

        if not default:
            default = 'h'
        for i, c in enumerate(choices):
            if isinstance(c, Separator):
                self.choices.append(c)
            else:
                if isinstance(c, basestring):
                    self.choices.append((key, c, c))
                else:
                    key = c.get('key')
                    name = c.get('name')
                    value = c.get('value', name)
                    if default and default == key:
                        self.pointer_index = i
                        key = key.upper()  # default key is in uppercase
                    self.choices.append((key, name, value))
        # append the help choice
        key = 'h'
        if not default:
            self.pointer_index = len(self.choices)
            key = key.upper()  # default key is in uppercase
        self.choices.append((key, 'Help, list all options', None))

    @property
    def choice_count(self):
        return len(self.choices)

    def _get_choice_tokens(self, cli):
        tokens = []
        T = Token

        def _append(index, line):
            if isinstance(line, Separator):
                tokens.append((T.Separator, '   %s\n' % line))
            else:
                key = line[0]
                line = line[1]
                pointed_at = (index == self.pointer_index)

                @if_mousedown
                def select_item(cli, mouse_event):
                    # bind option with this index to mouse event
                    self.pointer_index = index

                if pointed_at:
                    tokens.append((T.Selected, '  %s) %s' % (key, line),
                                   select_item))
                else:
                    tokens.append((T, '  %s) %s' % (key, line),
                                   select_item))
                tokens.append((T, '\n'))

        if self._help_active:
            # prepare the select choices
            for i, choice in enumerate(self.choices):
                _append(i, choice)
            tokens.append((T, '  Answer: %s' %
                           self.choices[self.pointer_index][0]))
        else:
            tokens.append((T.Pointer, '>> '))
            tokens.append((T, self.choices[self.pointer_index][1]))
        return tokens

    def get_selected_value(self):
        # get value not label
        return self.choices[self.pointer_index][2]


def question(message, **kwargs):
    # TODO extract common parts for list, checkbox, rawlist, expand
    # TODO up, down navigation
    if not 'choices' in kwargs:
        raise PromptParameterException('choices')

    choices = kwargs.pop('choices', None)
    default = kwargs.pop('default', None)

    # TODO style defaults on detail level
    style = kwargs.pop('style', default_style)

    ic = InquirerControl(choices, default)

    def get_prompt_tokens(cli):
        tokens = []
        T = Token

        tokens.append((T.QuestionMark, '?'))
        tokens.append((T.Question, ' %s ' % message))
        if not ic.answered:
            tokens.append((T.Instruction, ' (%s)' % ''.join(
                [k[0] for k in ic.choices if not isinstance(k, Separator)])))
        else:
            tokens.append((T.Answer, ' %s' % ic.get_selected_value()))
        return tokens

    #@Condition
    #def is_help_active(cli):
    #    return ic._help_active

    # assemble layout
    layout = HSplit([
        Window(height=D.exact(1),
               content=TokenListControl(get_prompt_tokens)
        ),
        ConditionalContainer(
            Window(ic),
            #filter=is_help_active & ~IsDone()  # ~ bitwise inverse
            filter=~IsDone()  # ~ bitwise inverse
        )
    ])

    # key bindings
    manager = KeyBindingManager.for_prompt()

    @manager.registry.add_binding(Keys.ControlQ, eager=True)
    @manager.registry.add_binding(Keys.ControlC, eager=True)
    def _(event):
        raise KeyboardInterrupt()

    # add key bindings for choices
    for i, c in enumerate(ic.choices):
        if not isinstance(c, Separator):
            def _reg_binding(i, keys):
                # trick out late evaluation with a "function factory":
                # http://stackoverflow.com/questions/3431676/creating-functions-in-a-loop
                @manager.registry.add_binding(keys, eager=True)
                def select_choice(event):
                    ic.pointer_index = i
            if c[0] not in ['h', 'H']:
                _reg_binding(i, c[0])
                if c[0].isupper():
                    _reg_binding(i, c[0].lower())

    @manager.registry.add_binding('H', eager=True)
    @manager.registry.add_binding('h', eager=True)
    def help_choice(event):
        ic._help_active = True

    @manager.registry.add_binding(Keys.Enter, eager=True)
    def set_answer(event):
        ic.answered = True
        event.cli.set_return_value(ic.get_selected_value())

    return Application(
        layout=layout,
        key_bindings_registry=manager.registry,
        mouse_support=True,
        style=style
    )



# -*- coding: utf-8 -*-
"""Tune hyperopt generation.
"""
from wandb.sweeps.config import cfg


class HyperOpt(cfg.SweepConfigElement):
    def __init__(self):
        super(HyperOpt, self).__init__()
            
    def HyperOptSearch(self,
                 space,
                 max_concurrent=None,
                 reward_attr=None,
                 metric=None,
                 mode=None,
                 points_to_evaluate=None,
                 n_initial_points=None,
                 random_state_seed=None,
                 gamma=None,
                 **kwargs):
        local_args = dict(locals())
        return self._config("hyperopt.HyperOptSearch", [], local_args)

HyperOptSearch = HyperOpt().HyperOptSearch





# -*- coding: utf-8 -*-
"""Tune engine.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six


class HyperOpt:
    def __init__(self):
        #super(Tune, self).__init__(_cfg_module, _cfg_version)
        pass

    def loguniform(self, config, version=None):
        from ray import tune
        args = dict(config)
        return tune.loguniform(**args)

    def uniform(self, config, version=None):
        from hyperopt import hp
        #print("got config", config)
        args = []
        kwargs = {}
        if isinstance(config, tuple):
            config = list(config)
        if isinstance(config, list):
            args = config
            if isinstance(args[-1], dict):
                kwargs = args.pop()
            # TODO: if final element is empty, drop it
        else:
            kwargs = config
        #print("about", args, kwargs)
        return lambda _: hp.uniform(*args, **kwargs)

    def choice(self, config, version=None):
        from hyperopt import hp
        #print("got config", config)
        args = []
        kwargs = {}
        if isinstance(config, tuple):
            config = list(config)
        if isinstance(config, list):
            args = config
            if isinstance(args[-1], dict):
                kwargs = args.pop()
            # TODO: if final element is empty, drop it
        else:
            kwargs = config
        #print("about", args, kwargs)
        #args = dict(config)
        return lambda _: hp.choice(*args, **kwargs)

    def randint(self, config, version=None):
        from ray import tune
        args = dict(config)
        return tune.randint(**args)

    def randn(self, config, version=None):
        from ray import tune
        args = dict(config)
        return tune.randn(**args)


hyperopt = HyperOpt()



# File is generated by: tox -e codemod
# -*- coding: utf-8 -*-
"""Backend Sender - Send to internal process

Manage backend sender.

"""

import json
import logging
import threading
import uuid

import six
from six.moves import queue
import wandb
from wandb import data_types
from wandb.proto import wandb_internal_pb2 as pb
from wandb.proto import wandb_telemetry_pb2 as tpb
from wandb.util import (
    get_h5_typename,
    json_dumps_safer,
    json_dumps_safer_history,
    json_friendly,
    json_friendly_val,
    maybe_compress_summary,
    WandBJSONEncoderOld,
)

from .artifacts import ArtifactManifest
from ..wandb_artifacts import Artifact

if wandb.TYPE_CHECKING:
    import typing as t
    from . import summary_record as sr
    from typing import Any, Dict, Iterable, Optional, Tuple, Union
    from multiprocessing import Process
    from typing import cast
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from ..wandb_run import Run
        from six.moves.queue import Queue
else:

    def cast(_, val):
        return val


logger = logging.getLogger("wandb")


def file_policy_to_enum(policy):
    if policy == "now":
        enum = pb.FilesItem.PolicyType.NOW
    elif policy == "end":
        enum = pb.FilesItem.PolicyType.END
    elif policy == "live":
        enum = pb.FilesItem.PolicyType.LIVE
    return enum


def file_enum_to_policy(enum):
    if enum == pb.FilesItem.PolicyType.NOW:
        policy = "now"
    elif enum == pb.FilesItem.PolicyType.END:
        policy = "end"
    elif enum == pb.FilesItem.PolicyType.LIVE:
        policy = "live"
    return policy


class _Future(object):
    # _object: Optional[pb.Result]

    def __init__(self):
        self._object = None
        self._object_ready = threading.Event()
        self._lock = threading.Lock()

    def get(self, timeout = None):
        is_set = self._object_ready.wait(timeout)
        if is_set and self._object:
            return self._object
        return None

    def _set_object(self, obj):
        self._object = obj
        self._object_ready.set()


class MessageRouter(object):
    # _pending_reqs: Dict[str, _Future]
    # _request_queue: "Queue[pb.Record]"
    # _response_queue: "Queue[pb.Result]"

    def __init__(
        self, request_queue, response_queue
    ):
        self._request_queue = request_queue
        self._response_queue = response_queue

        self._pending_reqs = {}
        self._lock = threading.Lock()

        self._join_event = threading.Event()
        self._thread = threading.Thread(target=self.message_loop)
        self._thread.daemon = True
        self._thread.start()

    def message_loop(self):
        while not self._join_event.is_set():
            try:
                msg = self._response_queue.get(timeout=1)
            except queue.Empty:
                continue
            self._handle_msg_rcv(msg)

    def send_and_receive(self, rec, local = None):
        rec.control.req_resp = True
        if local:
            rec.control.local = local
        rec.uuid = uuid.uuid4().hex
        future = _Future()
        with self._lock:
            self._pending_reqs[rec.uuid] = future

        self._request_queue.put(rec)

        return future

    def join(self):
        self._join_event.set()
        self._thread.join()

    def _handle_msg_rcv(self, msg):
        with self._lock:
            future = self._pending_reqs.pop(msg.uuid, None)
        if future is None:
            # TODO (cvp): saw this in tests, seemed benign enough to ignore, but
            # could point to other issues.
            if msg.uuid != "":
                logger.warning(
                    "No listener found for msg with uuid %s (%s)", msg.uuid, msg
                )
            return
        future._set_object(msg)


class BackendSender(object):
    class ExceptionTimeout(Exception):
        pass

    # record_q: Optional["Queue[pb.Record]"]
    # result_q: Optional["Queue[pb.Result]"]
    # process: Optional[Process]
    # _run: Optional["Run"]
    # _router: Optional[MessageRouter]

    def __init__(
        self,
        record_q = None,
        result_q = None,
        process = None,
    ):
        self.record_q = record_q
        self.result_q = result_q
        self._process = process
        self._run = None
        self._router = None

        if record_q and result_q:
            self._router = MessageRouter(record_q, result_q)

    def _hack_set_run(self, run):
        self._run = run

    def publish_output(self, name, data):
        # from vendor.protobuf import google3.protobuf.timestamp
        # ts = timestamp.Timestamp()
        # ts.GetCurrentTime()
        # now = datetime.now()
        if name == "stdout":
            otype = pb.OutputRecord.OutputType.STDOUT
        elif name == "stderr":
            otype = pb.OutputRecord.OutputType.STDERR
        else:
            # TODO(jhr): throw error?
            print("unknown type")
        o = pb.OutputRecord(output_type=otype, line=data)
        o.timestamp.GetCurrentTime()
        self._publish_output(o)

    def _publish_output(self, outdata):
        rec = pb.Record()
        rec.output.CopyFrom(outdata)
        self._publish(rec)

    def publish_tbdata(
        self, log_dir, save, root_logdir
    ):
        tbrecord = pb.TBRecord()
        tbrecord.log_dir = log_dir
        tbrecord.save = save
        tbrecord.root_dir = root_logdir or ""
        rec = self._make_record(tbrecord=tbrecord)
        self._publish(rec)

    def _publish_history(self, history):
        rec = self._make_record(history=history)
        self._publish(rec)

    def publish_history(
        self, data, step = None, run = None, publish_step = True
    ):
        run = run or self._run
        data = data_types.history_dict_to_json(run, data, step=step)
        history = pb.HistoryRecord()
        if publish_step:
            assert step is not None
            history.step.num = step
        data.pop("_step", None)
        for k, v in six.iteritems(data):
            item = history.item.add()
            item.key = k
            item.value_json = json_dumps_safer_history(v)  # type: ignore
        self._publish_history(history)

    def publish_telemetry(self, telem):
        rec = self._make_record(telemetry=telem)
        self._publish(rec)

    def _make_run(self, run):
        proto_run = pb.RunRecord()
        run._make_proto_run(proto_run)
        if run._settings.host:
            proto_run.host = run._settings.host
        if run._config is not None:
            config_dict = run._config._as_dict()  # type: ignore
            self._make_config(data=config_dict, obj=proto_run.config)
        if run._telemetry_obj:
            proto_run.telemetry.MergeFrom(run._telemetry_obj)
        return proto_run

    def _make_artifact(self, artifact):
        proto_artifact = pb.ArtifactRecord()
        proto_artifact.type = artifact.type
        proto_artifact.name = artifact.name
        proto_artifact.digest = artifact.digest
        if artifact.distributed_id:
            proto_artifact.distributed_id = artifact.distributed_id
        if artifact.description:
            proto_artifact.description = artifact.description
        if artifact.metadata:
            proto_artifact.metadata = json.dumps(json_friendly_val(artifact.metadata))  # type: ignore
        proto_artifact.incremental_beta1 = artifact.incremental
        self._make_artifact_manifest(artifact.manifest, obj=proto_artifact.manifest)
        return proto_artifact

    def _make_artifact_manifest(
        self, artifact_manifest, obj = None
    ):
        proto_manifest = obj or pb.ArtifactManifest()
        proto_manifest.version = artifact_manifest.version()  # type: ignore
        proto_manifest.storage_policy = artifact_manifest.storage_policy.name()

        for k, v in artifact_manifest.storage_policy.config().items() or {}.items():
            cfg = proto_manifest.storage_policy_config.add()
            cfg.key = k
            cfg.value_json = json.dumps(v)

        for entry in sorted(artifact_manifest.entries.values(), key=lambda k: k.path):  # type: ignore
            proto_entry = proto_manifest.contents.add()
            proto_entry.path = entry.path
            proto_entry.digest = entry.digest
            if entry.size:
                proto_entry.size = entry.size
            if entry.birth_artifact_id:
                proto_entry.birth_artifact_id = entry.birth_artifact_id
            if entry.ref:
                proto_entry.ref = entry.ref
            if entry.local_path:
                proto_entry.local_path = entry.local_path
            for k, v in entry.extra.items():
                proto_extra = proto_entry.extra.add()
                proto_extra.key = k
                proto_extra.value_json = json.dumps(v)
        return proto_manifest

    def _make_exit(self, exit_code):
        exit = pb.RunExitRecord()
        exit.exit_code = exit_code
        return exit

    def _make_config(
        self,
        data = None,
        key = None,
        val = None,
        obj = None,
    ):
        config = obj or pb.ConfigRecord()
        if data:
            for k, v in six.iteritems(data):
                update = config.update.add()
                update.key = k
                update.value_json = json_dumps_safer(json_friendly(v)[0])  # type: ignore
        if key:
            update = config.update.add()
            if isinstance(key, tuple):
                for k in key:
                    update.nested_key.append(k)
            else:
                update.key = key
            update.value_json = json_dumps_safer(json_friendly(val)[0])  # type: ignore
        return config

    def _make_stats(self, stats_dict):
        stats = pb.StatsRecord()
        stats.stats_type = pb.StatsRecord.StatsType.SYSTEM
        stats.timestamp.GetCurrentTime()
        for k, v in six.iteritems(stats_dict):
            item = stats.item.add()
            item.key = k
            item.value_json = json_dumps_safer(json_friendly(v)[0])  # type: ignore
        return stats

    def _summary_encode(self, value, path_from_root):
        """Normalize, compress, and encode sub-objects for backend storage.

        value: Object to encode.
        path_from_root: `str` dot separated string from the top-level summary to the
            current `value`.

        Returns:
            A new tree of dict's with large objects replaced with dictionaries
            with "_type" entries that say which type the original data was.
        """

        # Constructs a new `dict` tree in `json_value` that discards and/or
        # encodes objects that aren't JSON serializable.

        if isinstance(value, dict):
            json_value = {}
            for key, value in six.iteritems(value):
                json_value[key] = self._summary_encode(
                    value, path_from_root + "." + key
                )
            return json_value
        else:
            friendly_value, converted = json_friendly(  # type: ignore
                data_types.val_to_json(
                    self._run, path_from_root, value, namespace="summary"
                )
            )
            json_value, compressed = maybe_compress_summary(  # type: ignore
                friendly_value, get_h5_typename(value)  # type: ignore
            )
            if compressed:
                # TODO(jhr): impleement me
                pass
                # self.write_h5(path_from_root, friendly_value)

            return json_value

    def _make_summary_from_dict(self, summary_dict):
        summary = pb.SummaryRecord()
        for k, v in six.iteritems(summary_dict):
            update = summary.update.add()
            update.key = k
            update.value_json = json.dumps(v)
        return summary

    def _make_summary(self, summary_record):
        pb_summary_record = pb.SummaryRecord()

        for item in summary_record.update:
            pb_summary_item = pb_summary_record.update.add()
            key_length = len(item.key)

            assert key_length > 0

            if key_length > 1:
                pb_summary_item.nested_key.extend(item.key)
            else:
                pb_summary_item.key = item.key[0]

            path_from_root = ".".join(item.key)
            json_value = self._summary_encode(item.value, path_from_root)
            json_value, _ = json_friendly(json_value)  # type: ignore

            pb_summary_item.value_json = json.dumps(
                json_value, cls=WandBJSONEncoderOld,
            )

        for item in summary_record.remove:
            pb_summary_item = pb_summary_record.remove.add()
            key_length = len(item.key)

            assert key_length > 0

            if key_length > 1:
                pb_summary_item.nested_key.extend(item.key)
            else:
                pb_summary_item.key = item.key[0]

        return pb_summary_record

    def _make_files(self, files_dict):
        files = pb.FilesRecord()
        for path, policy in files_dict["files"]:
            f = files.files.add()
            f.path = path
            f.policy = file_policy_to_enum(policy)
        return files

    def _make_login(self, api_key = None):
        login = pb.LoginRequest()
        if api_key:
            login.api_key = api_key
        return login

    def _make_request(
        self,
        login = None,
        get_summary = None,
        pause = None,
        resume = None,
        stop_status = None,
        network_status = None,
        poll_exit = None,
        sampled_history = None,
        run_start = None,
        check_version = None,
        log_artifact = None,
        defer = None,
    ):
        request = pb.Request()
        if login:
            request.login.CopyFrom(login)
        elif get_summary:
            request.get_summary.CopyFrom(get_summary)
        elif pause:
            request.pause.CopyFrom(pause)
        elif resume:
            request.resume.CopyFrom(resume)
        elif stop_status:
            request.stop_status.CopyFrom(stop_status)
        elif network_status:
            request.network_status.CopyFrom(network_status)
        elif poll_exit:
            request.poll_exit.CopyFrom(poll_exit)
        elif sampled_history:
            request.sampled_history.CopyFrom(sampled_history)
        elif run_start:
            request.run_start.CopyFrom(run_start)
        elif check_version:
            request.check_version.CopyFrom(check_version)
        elif log_artifact:
            request.log_artifact.CopyFrom(log_artifact)
        elif defer:
            request.defer.CopyFrom(defer)
        else:
            raise Exception("Invalid request")
        record = self._make_record(request=request)
        # All requests do not get persisted
        record.control.local = True
        return record

    def _make_record(
        self,
        run = None,
        config = None,
        files = None,
        summary = None,
        history = None,
        stats = None,
        exit = None,
        artifact = None,
        tbrecord = None,
        alert = None,
        final = None,
        metric = None,
        header = None,
        footer = None,
        request = None,
        telemetry = None,
    ):
        record = pb.Record()
        if run:
            record.run.CopyFrom(run)
        elif config:
            record.config.CopyFrom(config)
        elif summary:
            record.summary.CopyFrom(summary)
        elif history:
            record.history.CopyFrom(history)
        elif files:
            record.files.CopyFrom(files)
        elif stats:
            record.stats.CopyFrom(stats)
        elif exit:
            record.exit.CopyFrom(exit)
        elif artifact:
            record.artifact.CopyFrom(artifact)
        elif tbrecord:
            record.tbrecord.CopyFrom(tbrecord)
        elif alert:
            record.alert.CopyFrom(alert)
        elif final:
            record.final.CopyFrom(final)
        elif header:
            record.header.CopyFrom(header)
        elif footer:
            record.footer.CopyFrom(footer)
        elif request:
            record.request.CopyFrom(request)
        elif telemetry:
            record.telemetry.CopyFrom(telemetry)
        elif metric:
            record.metric.CopyFrom(metric)
        else:
            raise Exception("Invalid record")
        return record

    def _publish(self, record, local = None):
        if self._process and not self._process.is_alive():
            raise Exception("The wandb backend process has shutdown")
        if local:
            record.control.local = local
        if self.record_q:
            self.record_q.put(record)

    def _communicate(
        self, rec, timeout = 5, local = None
    ):
        return self._communicate_async(rec, local=local).get(timeout=timeout)

    def _communicate_async(self, rec, local = None):
        assert self._router
        if self._process and not self._process.is_alive():
            raise Exception("The wandb backend process has shutdown")
        future = self._router.send_and_receive(rec, local=local)
        return future

    def communicate_login(
        self, api_key = None, timeout = 15
    ):
        login = self._make_login(api_key)
        rec = self._make_request(login=login)
        result = self._communicate(rec, timeout=timeout)
        if result is None:
            # TODO: friendlier error message here
            raise wandb.Error(
                "Couldn't communicate with backend after %s seconds" % timeout
            )
        login_response = result.response.login_response
        assert login_response
        return login_response

    def _publish_defer(self, state):
        defer = pb.DeferRequest(state=state)
        rec = self._make_request(defer=defer)
        self._publish(rec, local=True)

    def publish_defer(self, state = 0):
        self._publish_defer(cast("pb.DeferRequest.DeferStateValue", state))

    def publish_header(self):
        header = pb.HeaderRecord()
        rec = self._make_record(header=header)
        self._publish(rec)

    def publish_footer(self):
        footer = pb.FooterRecord()
        rec = self._make_record(footer=footer)
        self._publish(rec)

    def publish_final(self):
        final = pb.FinalRecord()
        rec = self._make_record(final=final)
        self._publish(rec)

    def publish_login(self, api_key = None):
        login = self._make_login(api_key)
        rec = self._make_request(login=login)
        self._publish(rec)

    def publish_pause(self):
        pause = pb.PauseRequest()
        rec = self._make_request(pause=pause)
        self._publish(rec)

    def publish_resume(self):
        resume = pb.ResumeRequest()
        rec = self._make_request(resume=resume)
        self._publish(rec)

    def _publish_run(self, run):
        rec = self._make_record(run=run)
        self._publish(rec)

    def publish_run(self, run_obj):
        run = self._make_run(run_obj)
        self._publish_run(run)

    def publish_config(
        self,
        data = None,
        key = None,
        val = None,
    ):
        cfg = self._make_config(data=data, key=key, val=val)

        self._publish_config(cfg)

    def _publish_config(self, cfg):
        rec = self._make_record(config=cfg)
        self._publish(rec)

    def publish_summary(self, summary_record):
        pb_summary_record = self._make_summary(summary_record)
        self._publish_summary(pb_summary_record)

    def _publish_summary(self, summary):
        rec = self._make_record(summary=summary)
        self._publish(rec)

    def _publish_metric(self, metric):
        rec = self._make_record(metric=metric)
        self._publish(rec)

    def _communicate_run(
        self, run, timeout = None
    ):
        """Send synchronous run object waiting for a response.

        Arguments:
            run: RunRecord object
            timeout: number of seconds to wait

        Returns:
            RunRecord object
        """

        req = self._make_record(run=run)
        resp = self._communicate(req, timeout=timeout)
        if resp is None:
            logger.info("couldn't get run from backend")
            # Note: timeouts handled by callers: wandb_init.py
            return None
        assert resp.HasField("run_result")
        return resp.run_result

    def communicate_run(
        self, run_obj, timeout = None
    ):
        run = self._make_run(run_obj)
        return self._communicate_run(run, timeout=timeout)

    def publish_stats(self, stats_dict):
        stats = self._make_stats(stats_dict)
        rec = self._make_record(stats=stats)
        self._publish(rec)

    def publish_files(self, files_dict):
        files = self._make_files(files_dict)
        rec = self._make_record(files=files)
        self._publish(rec)

    def communicate_artifact(
        self,
        run,
        artifact,
        aliases,
        is_user_created = False,
        use_after_commit = False,
        finalize = True,
    ):
        proto_run = self._make_run(run)
        proto_artifact = self._make_artifact(artifact)
        proto_artifact.run_id = proto_run.run_id
        proto_artifact.project = proto_run.project
        proto_artifact.entity = proto_run.entity
        proto_artifact.user_created = is_user_created
        proto_artifact.use_after_commit = use_after_commit
        proto_artifact.finalize = finalize
        for alias in aliases:
            proto_artifact.aliases.append(alias)

        log_artifact = pb.LogArtifactRequest()
        log_artifact.artifact.CopyFrom(proto_artifact)
        rec = self._make_request(log_artifact=log_artifact)
        return self._communicate_async(rec)

    def publish_artifact(
        self,
        run,
        artifact,
        aliases,
        is_user_created = False,
        use_after_commit = False,
        finalize = True,
    ):
        proto_run = self._make_run(run)
        proto_artifact = self._make_artifact(artifact)
        proto_artifact.run_id = proto_run.run_id
        proto_artifact.project = proto_run.project
        proto_artifact.entity = proto_run.entity
        proto_artifact.user_created = is_user_created
        proto_artifact.use_after_commit = use_after_commit
        proto_artifact.finalize = finalize
        for alias in aliases:
            proto_artifact.aliases.append(alias)
        rec = self._make_record(artifact=proto_artifact)
        self._publish(rec)

    def publish_alert(
        self, title, text, level, wait_duration
    ):
        proto_alert = pb.AlertRecord()
        proto_alert.title = title
        proto_alert.text = text
        proto_alert.level = level
        proto_alert.wait_duration = wait_duration
        rec = self._make_record(alert=proto_alert)
        self._publish(rec)

    def communicate_stop_status(
        self, timeout = None
    ):
        status = pb.StopStatusRequest()
        req = self._make_request(stop_status=status)

        resp = self._communicate(req, timeout=timeout, local=True)
        if resp is None:
            return None
        assert resp.response.stop_status_response
        return resp.response.stop_status_response

    def communicate_network_status(
        self, timeout = None
    ):
        status = pb.NetworkStatusRequest()
        req = self._make_request(network_status=status)

        resp = self._communicate(req, timeout=timeout, local=True)
        if resp is None:
            return None
        assert resp.response.network_status_response
        return resp.response.network_status_response

    def publish_exit(self, exit_code):
        exit_data = self._make_exit(exit_code)
        rec = self._make_record(exit=exit_data)
        self._publish(rec)

    def _communicate_exit(
        self, exit_data, timeout = None
    ):
        req = self._make_record(exit=exit_data)

        result = self._communicate(req, timeout=timeout)
        if result is None:
            # TODO: friendlier error message here
            raise wandb.Error(
                "Couldn't communicate with backend after %s seconds" % timeout
            )
        assert result.exit_result
        return result.exit_result

    def communicate_poll_exit(self):
        poll_request = pb.PollExitRequest()
        rec = self._make_request(poll_exit=poll_request)
        result = self._communicate(rec)
        if result is None:
            return None
        poll_exit_response = result.response.poll_exit_response
        assert poll_exit_response
        return poll_exit_response

    def communicate_check_version(
        self, current_version = None
    ):
        check_version = pb.CheckVersionRequest()
        if current_version:
            check_version.current_version = current_version
        rec = self._make_request(check_version=check_version)
        result = self._communicate(rec)
        if result is None:
            # Note: timeouts handled by callers: wandb_init.py
            return None
        return result.response.check_version_response

    def communicate_run_start(self, run_pb):
        run_start = pb.RunStartRequest()
        run_start.run.CopyFrom(run_pb)
        rec = self._make_request(run_start=run_start)
        result = self._communicate(rec)
        return result

    def communicate_exit(self, exit_code, timeout = None):
        exit_data = self._make_exit(exit_code)
        return self._communicate_exit(exit_data, timeout=timeout)

    def communicate_summary(self):
        record = self._make_request(get_summary=pb.GetSummaryRequest())
        result = self._communicate(record, timeout=10)
        if result is None:
            return None
        get_summary_response = result.response.get_summary_response
        assert get_summary_response
        return get_summary_response

    def communicate_sampled_history(self):
        record = self._make_request(sampled_history=pb.SampledHistoryRequest())
        result = self._communicate(record)
        if result is None:
            return None
        sampled_history_response = result.response.sampled_history_response
        assert sampled_history_response
        return sampled_history_response

    def join(self):
        # shutdown
        request = pb.Request(shutdown=pb.ShutdownRequest())
        record = self._make_record(request=request)
        _ = self._communicate(record)

        if self._router:
            self._router.join()



from __future__ import unicode_literals

import datetime
import os
import subprocess


def get_version(version=None):
    "Returns a PEP 440-compliant version number from VERSION."
    version = get_complete_version(version)

    # Now build the two parts of the version number:
    # main = X.Y[.Z]
    # sub = .devN - for pre-alpha releases
    #     | {a|b|rc}N - for alpha, beta, and rc releases

    main = get_main_version(version)

    sub = ''
    if version[3] == 'alpha' and version[4] == 0:
        git_changeset = get_git_changeset()
        if git_changeset:
            sub = '.dev%s' % git_changeset
        else:
            sub = '.dev'
    elif version[3] != 'final':
        mapping = {'alpha': 'a', 'beta': 'b', 'rc': 'rc'}
        sub = mapping[version[3]] + str(version[4])

    return str(main + sub)


def get_main_version(version=None):
    "Returns main version (X.Y[.Z]) from VERSION."
    version = get_complete_version(version)
    parts = 2 if version[2] == 0 else 3
    return '.'.join(str(x) for x in version[:parts])


def get_complete_version(version=None):
    """Returns a tuple of the graphql version. If version argument is non-empty,
    then checks for correctness of the tuple provided.
    """
    if version is None:
        from graphql import VERSION as version
    else:
        assert len(version) == 5
        assert version[3] in ('alpha', 'beta', 'rc', 'final')

    return version


def get_docs_version(version=None):
    version = get_complete_version(version)
    if version[3] != 'final':
        return 'dev'
    else:
        return '%d.%d' % version[:2]


def get_git_changeset():
    """Returns a numeric identifier of the latest git changeset.
    The result is the UTC timestamp of the changeset in YYYYMMDDHHMMSS format.
    This value isn't guaranteed to be unique, but collisions are very unlikely,
    so it's sufficient for generating the development version numbers.
    """
    repo_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    try:
        git_log = subprocess.Popen(
            'git log --pretty=format:%ct --quiet -1 HEAD',
            stdout=subprocess.PIPE, stderr=subprocess.PIPE,
            shell=True, cwd=repo_dir, universal_newlines=True,
        )
        timestamp = git_log.communicate()[0]
        timestamp = datetime.datetime.utcfromtimestamp(int(timestamp))
    except:
        return None
    return timestamp.strftime('%Y%m%d%H%M%S')



# File is generated by: tox -e codemod


class RunDisabled(str):
    def __init__(self, *args, **kwargs):
        object.__setattr__(self, "___dict", {})

    def __add__(self, other):
        return self

    def __sub__(self, other):
        return self

    def __mul__(self, other):
        return self

    def __truediv__(self, other):
        return self

    def __floordiv__(self, other):
        return self

    def __mod__(self, other):
        return self

    def __pow__(self, other, modulo=None):
        return self

    def __lshift__(self, other):
        return self

    def __rshift__(self, other):
        return self

    def __and__(self, other):
        return self

    def __xor__(self, other):
        return self

    def __or__(self, other):
        return self

    def __iadd__(self, other):
        return self

    def __isub__(self, other):
        return self

    def __imul__(self, other):
        return self

    def __idiv__(self, other):
        return self

    def __ifloordiv__(self, other):
        return self

    def __imod__(self, other):
        return self

    def __ipow__(self, other, modulo=None):
        return self

    def __ilshift__(self, other):
        return self

    def __irshift__(self, other):
        return self

    def __iand__(self, other):
        return self

    def __ixor__(self, other):
        return self

    def __ior__(self, other):
        return self

    def __neg__(self):
        return self

    def __pos__(self):
        return self

    def __abs__(self):
        return self

    def __invert__(self):
        return self

    def __complex__(self):
        return 1 + 0j

    def __int__(self):
        return 1

    def __long__(self):
        return 1

    def __float__(self):
        return 1.0

    def __oct__(self):
        return oct(1)

    def __hex__(self):
        return hex(1)

    def __lt__(self, other):
        return True

    def __le__(self, other):
        return True

    def __eq__(self, other):
        return True

    def __ne__(self, other):
        return True

    def __gt__(self, other):
        return True

    def __ge__(self, other):
        return True

    def __getattr__(self, attr):
        return self[attr]

    def __getitem__(self, key):
        d = object.__getattribute__(self, "___dict")
        try:
            if key in d:
                return d[key]
        except TypeError:
            key = str(key)
            if key in d:
                return d[key]
        dummy = RunDisabled()
        d[key] = dummy
        return dummy

    def __setitem__(self, key, value):
        object.__getattribute__(self, "___dict")[key] = value

    def __setattr__(self, key, value):
        self[key] = value

    def __call__(self, *args, **kwargs):
        return RunDisabled()

    def __len__(self):
        return 1

    def __str__(self):
        return ""

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return exc_type is None

    def __repr__(self):
        return ""

    def __nonzero__(self):
        return True

    def __bool__(self):
        return True

    def __getstate__(self):
        return 1


class SummaryDisabled(dict):
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __getattr__(self, key):
        return self[key]

    def __getitem__(self, key):
        val = dict.__getitem__(self, key)
        if isinstance(val, dict) and not isinstance(val, SummaryDisabled):
            val = SummaryDisabled(val)
            self[key] = val
        return val



# -*- coding: utf-8 -*-
"""Sweep config interface.

Example:
    from wandb.sweeps import configure as cfg
    ray_config = cfg.tune.run(
        search_algo = cfg.tune.suggest.hyperopt.HyperOptSearch(
            space = {
                'width': cfg.hp.uniform('width', 0, 20),
                'height': cfg.hp.uniform('height', -100, 100),
                'activation': cfg.hp.choice("activation", ["relu", "tanh"])
            },
            max_concurrent = 4,
            reward_attr = "neg_mean_loss"
        ),
        scheduler = cfg.tune.schedulers.AsyncHyperBandScheduler(reward_attr="neg_mean_loss"),
        num_samples = 10 if args.smoke_test else 1000,
        config = {
            "iterations": 100,
        },
        stop = {
            "timesteps_total": 100
        },
    )
    id = wandb.sweep(ray_config)
    print("id:", id)
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

#from wandb.sweeps.config.tune import tune
from wandb.sweeps.config.cfg import SweepConfig
#from wandb.sweeps.config import hyperopt

__all__ = [
    "tune",
    "SweepConfig",
]




#!/usr/bin/env python

"""All of W&B's environment variables

Getters and putters for all of them should go here. That
way it'll be easier to avoid typos with names and be
consistent about environment variables' semantics.

Environment variables are not the authoritative source for
these values in many cases.
"""

import os
import sys
import json
import wandb
from distutils.util import strtobool

CONFIG_PATHS = "WANDB_CONFIG_PATHS"
SWEEP_PARAM_PATH = "WANDB_SWEEP_PARAM_PATH"
SHOW_RUN = "WANDB_SHOW_RUN"
DEBUG = "WANDB_DEBUG"
SILENT = "WANDB_SILENT"
INITED = "WANDB_INITED"
DIR = "WANDB_DIR"
# Deprecate DESCRIPTION in a future release
DESCRIPTION = "WANDB_DESCRIPTION"
NAME = "WANDB_NAME"
NOTEBOOK_NAME = "WANDB_NOTEBOOK_NAME"
NOTES = "WANDB_NOTES"
USERNAME = "WANDB_USERNAME"
USER_EMAIL = "WANDB_USER_EMAIL"
PROJECT = "WANDB_PROJECT"
ENTITY = "WANDB_ENTITY"
BASE_URL = "WANDB_BASE_URL"
PROGRAM = "WANDB_PROGRAM"
ARGS = "WANDB_ARGS"
MODE = "WANDB_MODE"
START_METHOD = "WANDB_START_METHOD"
RESUME = "WANDB_RESUME"
RUN_ID = "WANDB_RUN_ID"
RUN_STORAGE_ID = "WANDB_RUN_STORAGE_ID"
RUN_GROUP = "WANDB_RUN_GROUP"
RUN_DIR = "WANDB_RUN_DIR"
SWEEP_ID = "WANDB_SWEEP_ID"
HTTP_TIMEOUT = "WANDB_HTTP_TIMEOUT"
API_KEY = "WANDB_API_KEY"
JOB_TYPE = "WANDB_JOB_TYPE"
DISABLE_CODE = "WANDB_DISABLE_CODE"
SAVE_CODE = "WANDB_SAVE_CODE"
TAGS = "WANDB_TAGS"
IGNORE = "WANDB_IGNORE_GLOBS"
ERROR_REPORTING = "WANDB_ERROR_REPORTING"
DOCKER = "WANDB_DOCKER"
AGENT_REPORT_INTERVAL = "WANDB_AGENT_REPORT_INTERVAL"
AGENT_KILL_DELAY = "WANDB_AGENT_KILL_DELAY"
AGENT_DISABLE_FLAPPING = "WANDB_AGENT_DISABLE_FLAPPING"
AGENT_MAX_INITIAL_FAILURES = "WANDB_AGENT_MAX_INITIAL_FAILURES"
CRASH_NOSYNC_TIME = "WANDB_CRASH_NOSYNC_TIME"
MAGIC = "WANDB_MAGIC"
HOST = "WANDB_HOST"
ANONYMOUS = "WANDB_ANONYMOUS"
JUPYTER = "WANDB_JUPYTER"
CONFIG_DIR = "WANDB_CONFIG_DIR"
CACHE_DIR = "WANDB_CACHE_DIR"

# For testing, to be removed in future version
USE_V1_ARTIFACTS = "_WANDB_USE_V1_ARTIFACTS"


def immutable_keys():
    """These are env keys that shouldn't change within a single process.  We use this to maintain
    certain values between multiple calls to wandb.init within a single process."""
    return [
        DIR,
        ENTITY,
        PROJECT,
        API_KEY,
        IGNORE,
        DISABLE_CODE,
        DOCKER,
        MODE,
        BASE_URL,
        ERROR_REPORTING,
        CRASH_NOSYNC_TIME,
        MAGIC,
        USERNAME,
        USER_EMAIL,
        DIR,
        SILENT,
        CONFIG_PATHS,
        ANONYMOUS,
        RUN_GROUP,
        JOB_TYPE,
        TAGS,
        RESUME,
        AGENT_REPORT_INTERVAL,
        HTTP_TIMEOUT,
        HOST,
        CACHE_DIR,
        USE_V1_ARTIFACTS,
    ]


def _env_as_bool(var, default=None, env=None):
    if env is None:
        env = os.environ
    val = env.get(var, default)
    try:
        val = bool(strtobool(val))
    except (AttributeError, ValueError):
        pass
    return val if isinstance(val, bool) else False


def is_debug(default=None, env=None):
    return _env_as_bool(DEBUG, default=default, env=env)


def error_reporting_enabled():
    return _env_as_bool(ERROR_REPORTING, default=True)


def get_error_reporting(default=True, env=None):
    if env is None:
        env = os.environ

    return env.get(ERROR_REPORTING, default)


def get_run(default=None, env=None):
    if env is None:
        env = os.environ

    return env.get(RUN_ID, default)


def get_args(default=None, env=None):
    if env is None:
        env = os.environ
    if env.get(ARGS):
        try:
            return json.loads(env.get(ARGS, "[]"))
        except ValueError:
            return None
    else:
        return default or sys.argv[1:]


def get_docker(default=None, env=None):
    if env is None:
        env = os.environ

    return env.get(DOCKER, default)


def get_http_timeout(default=10, env=None):
    if env is None:
        env = os.environ

    return int(env.get(HTTP_TIMEOUT, default))


def get_ignore(default=None, env=None):
    if env is None:
        env = os.environ

    if env.get(IGNORE, default):
        return env.get(IGNORE, default).split(",")
    else:
        return []


def get_project(default=None, env=None):
    if env is None:
        env = os.environ

    return env.get(PROJECT, default)


def get_username(default=None, env=None):
    if env is None:
        env = os.environ

    return env.get(USERNAME, default)


def get_user_email(default=None, env=None):
    if env is None:
        env = os.environ

    return env.get(USER_EMAIL, default)


def get_entity(default=None, env=None):
    if env is None:
        env = os.environ

    return env.get(ENTITY, default)


def get_base_url(default=None, env=None):
    if env is None:
        env = os.environ

    return env.get(BASE_URL, default)


def get_show_run(default=None, env=None):
    if env is None:
        env = os.environ

    return bool(env.get(SHOW_RUN, default))


def get_description(default=None, env=None):
    if env is None:
        env = os.environ

    return env.get(DESCRIPTION, default)


def get_tags(default="", env=None):
    if env is None:
        env = os.environ

    return [tag for tag in env.get(TAGS, default).split(",") if tag]


def get_dir(default=None, env=None):
    if env is None:
        env = os.environ
    return env.get(DIR, default)


def get_config_paths(default=None, env=None):
    if env is None:
        env = os.environ
    return env.get(CONFIG_PATHS, default)


def get_agent_report_interval(default=None, env=None):
    if env is None:
        env = os.environ
    val = env.get(AGENT_REPORT_INTERVAL, default)
    try:
        val = int(val)
    except ValueError:
        val = None  # silently ignore env format errors, caller should handle.
    return val


def get_agent_kill_delay(default=None, env=None):
    if env is None:
        env = os.environ
    val = env.get(AGENT_KILL_DELAY, default)
    try:
        val = int(val)
    except ValueError:
        val = None  # silently ignore env format errors, caller should handle.
    return val


def get_crash_nosync_time(default=None, env=None):
    if env is None:
        env = os.environ
    val = env.get(CRASH_NOSYNC_TIME, default)
    try:
        val = int(val)
    except ValueError:
        val = None  # silently ignore env format errors, caller should handle.
    return val


def get_magic(default=None, env=None):
    if env is None:
        env = os.environ
    val = env.get(MAGIC, default)
    return val


def get_cache_dir(env=None):
    default_dir = os.path.expanduser(os.path.join("~", ".cache", "wandb"))
    if env is None:
        env = os.environ
    val = env.get(CACHE_DIR, default_dir)
    return val


def get_use_v1_artifacts(env=None):
    if env is None:
        env = os.environ
    val = env.get(USE_V1_ARTIFACTS, False)
    return val


def get_agent_max_initial_failures(default=None, env=None):
    if env is None:
        env = os.environ
    val = env.get(AGENT_MAX_INITIAL_FAILURES, default)
    try:
        val = int(val)
    except ValueError:
        val = default
    return val


def set_entity(value, env=None):
    if env is None:
        env = os.environ
    env[ENTITY] = value


def set_project(value, env=None):
    if env is None:
        env = os.environ
    env[PROJECT] = value or "uncategorized"


def should_save_code():
    save_code = _env_as_bool(SAVE_CODE, default=False)
    code_disabled = _env_as_bool(DISABLE_CODE, default=False)
    # SAVE_CODE takes precedence over DISABLE_CODE
    return save_code and not code_disabled



#!/usr/bin/env python
"""WIP wandb grpc server."""

from concurrent import futures
import datetime
import logging
import multiprocessing
import os
import time

import grpc
import wandb
from wandb import wandb_sdk
from wandb.proto import wandb_internal_pb2  # type: ignore
from wandb.proto import wandb_server_pb2  # type: ignore
from wandb.proto import wandb_server_pb2_grpc  # type: ignore


class InternalServiceServicer(wandb_server_pb2_grpc.InternalServiceServicer):
    """Provides methods that implement functionality of route guide server."""

    def __init__(self, server, backend):
        self._server = server
        self._backend = backend

    def RunUpdate(self, run_data, context):  # noqa: N802
        if not run_data.run_id:
            run_data.run_id = wandb_sdk.lib.runid.generate_id()
        # Record telemetry info about grpc server
        run_data.telemetry.feature.grpc = True
        run_data.telemetry.cli_version = wandb.__version__
        result = self._backend._interface._communicate_run(run_data)

        # initiate run (stats and metadata probing)
        _ = self._backend._interface.communicate_run_start(result.run)

        return result

    def RunExit(self, exit_data, context):  # noqa: N802
        result = self._backend._interface._communicate_exit(exit_data)
        return result

    def Log(self, log_data, context):  # noqa: N802
        # TODO: make this sync?
        self._backend._interface._publish_history(log_data)
        # make up a response even though this was async
        result = wandb_internal_pb2.HistoryResult()
        return result

    def Summary(self, summary_data, context):  # noqa: N802
        # TODO: make this sync?
        self._backend._interface._publish_summary(summary_data)
        # make up a response even though this was async
        result = wandb_internal_pb2.SummaryResult()
        return result

    def Output(self, output_data, context):  # noqa: N802
        # TODO: make this sync?
        self._backend._interface._publish_output(output_data)
        # make up a response even though this was async
        result = wandb_internal_pb2.OutputResult()
        return result

    def Config(self, config_data, context):  # noqa: N802
        # TODO: make this sync?
        self._backend._interface._publish_config(config_data)
        # make up a response even though this was async
        result = wandb_internal_pb2.ConfigResult()
        return result

    def ServerShutdown(self, request, context):  # noqa: N802
        self._backend.cleanup()
        result = wandb_server_pb2.ServerShutdownResult()
        self._server.stop(5)
        return result

    def ServerStatus(self, request, context):  # noqa: N802
        result = wandb_server_pb2.ServerStatusResult()
        return result


# TODO(jhr): this should be merged with code in backend/backend.py ensure launched
class Backend:
    def __init__(self):
        self._done = False
        self._interface = None
        self._record_q = None
        self._result_q = None

    def setup(self):
        log_level = logging.DEBUG
        start_time = time.time()
        start_datetime = datetime.datetime.now()
        timespec = datetime.datetime.strftime(start_datetime, "%Y%m%d_%H%M%S")

        wandb_dir = "wandb"
        run_path = "run-{}-server".format(timespec)
        run_dir = os.path.join(wandb_dir, run_path)
        files_dir = os.path.join(run_dir, "files")
        sync_file = os.path.join(run_dir, "run-{}.wandb".format(start_time))
        os.makedirs(files_dir)
        settings = dict(
            log_internal=os.path.join(run_dir, "internal.log"),
            files_dir=files_dir,
            _start_time=start_time,
            _start_datetime=start_datetime,
            disable_code=None,
            code_program=None,
            save_code=None,
            sync_file=sync_file,
            _internal_queue_timeout=20,
            _internal_check_process=0,
            _disable_meta=True,
            _disable_stats=False,
            git_remote=None,
            program=None,
            resume=None,
            ignore_globs=(),
            offline=None,
            _log_level=log_level,
            run_id=None,
            entity=None,
            project=None,
            run_group=None,
            run_job_type=None,
            run_tags=None,
            run_name=None,
            run_notes=None,
            _jupyter=None,
            _kaggle=None,
            _offline=None,
            email=None,
            silent=None,
        )

        mp = multiprocessing
        fd_pipe_child, fd_pipe_parent = mp.Pipe()

        record_q = mp.Queue()
        # TODO: should this be one item just to make sure it stays fully synchronous?
        result_q = mp.Queue()

        wandb_process = mp.Process(
            target=wandb_sdk.internal.internal.wandb_internal,
            kwargs=dict(settings=settings, record_q=record_q, result_q=result_q,),
        )
        wandb_process.name = "wandb_internal"
        wandb_process.start()

        self.record_q = record_q
        self.result_q = result_q
        self.wandb_process = wandb_process

        self._interface = wandb_sdk.interface.interface.BackendSender(
            record_q=record_q, result_q=result_q, process=wandb_process,
        )

    def cleanup(self):
        # TODO: make _done atomic
        if self._done:
            return
        self._done = True
        self._interface.join()
        self.wandb_process.join()
        self.record_q.close()
        self.result_q.close()
        # No printing allowed from here until redirect restore!!!


def serve(backend, port):
    try:
        server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
        wandb_server_pb2_grpc.add_InternalServiceServicer_to_server(
            InternalServiceServicer(server, backend), server
        )
        server.add_insecure_port("[::]:{}".format(port))
        server.start()
        server.wait_for_termination()
        # print("server shutting down")
        # print("shutdown")
    except KeyboardInterrupt:
        print("control-c")


def main(port=None):
    try:
        logging.basicConfig()
        backend = Backend()
        backend.setup()
        serve(backend, port or 50051)
    except KeyboardInterrupt:
        print("outer control-c")


if __name__ == "__main__":
    main()



# File is generated by: tox -e codemod
from enum import Enum

"""
Call run.alert() to generate an email or Slack notification programmatically.
"""


class AlertLevel(Enum):
    INFO = "INFO"
    WARN = "WARN"
    ERROR = "ERROR"



"""W&B callback for lightgbm

Really simple callback to get logging for each tree

Example usage:

param_list = [("eta", 0.08), ("max_depth", 6), ("subsample", 0.8), ("colsample_bytree", 0.8), ("alpha", 8), ("num_class", 10)]
config.update(dict(param_list))
lgb = lgb.train(param_list, d_train, callbacks=[wandb_callback()])
"""

import lightgbm
import wandb


def wandb_callback():
    def callback(env):
        eval_results = {}
        recorder = lightgbm.record_evaluation(eval_results)
        recorder(env)

        for validation_key in eval_results.keys():
            for key in eval_results[validation_key].keys():
                wandb.log(
                    {validation_key + "_" + key: eval_results[validation_key][key][0]},
                    commit=False,
                )
        # Previous log statements use commit=False. This commits them.
        wandb.log({})

    return callback



"""
Sweep service
"""

import itertools
import random
import numpy as np
import math
from wandb.sweeps import grid_search, bayes_search, random_search
from wandb.sweeps import raytune
from wandb.sweeps import hyperband_stopping, envelope_stopping
from wandb.sweeps import base


class Search(base.Search):
    @staticmethod
    def to_class(config):
        if config.get('tune'):
            return raytune.RayTuneSearch()
        method = config.get('method')
        if method is None:
            raise ValueError('config missing required "method" field.')
        method = method.lower()
        if method == 'grid':
            return grid_search.GridSearch()
        elif method == 'bayes':
            return bayes_search.BayesianSearch()
        elif method == 'random':
            return random_search.RandomSearch()
        raise ValueError('method "%s" is not supported' % config['method'])


class EarlyTerminate(base.EarlyTerminate):
    @staticmethod
    def to_class(config):
        et_config = config.get('early_terminate')
        if not et_config:
            return base.EarlyTerminate()
        et_type = et_config.get('type')
        if not et_type:
            raise ValueError("Didn't specify early terminate type")
        et_type = et_type.lower()
        if et_type == 'envelope':
            return envelope_stopping.EnvelopeEarlyTerminate.init_from_config(et_config)
        elif et_type == 'hyperband':
            return hyperband_stopping.HyperbandEarlyTerminate.init_from_config(et_config)
        raise ValueError(
            'unsupported early termination type %s' % et_type)



# -*- coding: utf-8 -*-

from __future__ import absolute_import, print_function
import os

from prompt_toolkit.token import Token
from prompt_toolkit.styles import style_from_dict
from prompt_toolkit.validation import Validator, ValidationError

from .utils import print_json, format_json


__version__ = '0.5.2'


def here(p):
    return os.path.abspath(os.path.join(os.path.dirname(__file__), p))


class PromptParameterException(ValueError):
    def __init__(self, message, errors=None):

        # Call the base class constructor with the parameters it needs
        super(PromptParameterException, self).__init__(
            'You must provide a `%s` value' % message, errors)

from .prompt import prompt
from .separator import Separator
from .prompts.common import default_style



"""
Bayesian Search

Check out https://arxiv.org/pdf/1206.2944.pdf
 for explanation of bayesian optimization

We do bayesian optimization and handle the cases where some X values are integers
as well as the case where X is very large.

"""

import numpy as np
#from sklearn.gaussian_process import GaussianProcessRegressor
#from sklearn.gaussian_process.kernels import Matern
#import scipy.stats as stats
import math
from wandb.util import get_module
from wandb.sweeps.base import Search
from wandb.sweeps.params import HyperParameter, HyperParameterSet

sklearn_gaussian = get_module('sklearn.gaussian_process')
scipy_stats = get_module('scipy.stats')


def fit_normalized_gaussian_process(X, y, nu=1.5):
    """
        We fit a gaussian process but first subtract the mean and divide by stddev.
        To undo at prediction tim, call y_pred = gp.predict(X) * y_stddev + y_mean
    """
    gp = sklearn_gaussian.GaussianProcessRegressor(
        kernel=sklearn_gaussian.kernels.Matern(nu=nu), n_restarts_optimizer=2, alpha=0.0000001, random_state=2
    )
    if len(y) == 1:
        y = np.array(y)
        y_mean = y[0]
        y_stddev = 1
    else:
        y_mean = np.mean(y)
        y_stddev = np.std(y) + 0.0001
    y_norm = (y - y_mean) / y_stddev
    gp.fit(X, y_norm)
    return gp, y_mean, y_stddev


def sigmoid(x):
    return np.exp(-np.logaddexp(0, -x))


def random_sample(X_bounds, num_test_samples):
    num_hyperparameters = len(X_bounds)
    test_X = np.empty((num_test_samples, num_hyperparameters))
    for ii in range(num_test_samples):
        for jj in range(num_hyperparameters):
            if type(X_bounds[jj][0]) == int:
                assert (type(X_bounds[jj][1]) == int)
                test_X[ii, jj] = np.random.randint(
                    X_bounds[jj][0], X_bounds[jj][1])
            else:
                test_X[ii, jj] = np.random.uniform() * (
                    X_bounds[jj][1] - X_bounds[jj][0]
                ) + X_bounds[
                    jj
                ][
                    0
                ]
    return test_X


def predict(X, y, test_X, nu=1.5):
    gp, norm_mean, norm_stddev = fit_normalized_gaussian_process(X, y, nu=nu)
    y_pred, y_std = gp.predict([test_X], return_std=True)
    y_std_norm = y_std * norm_stddev
    y_pred_norm = (y_pred * norm_stddev) + norm_mean
    return y_pred_norm[0], y_std_norm[0]


def train_runtime_model(sample_X, runtimes, X_bounds):
    if sample_X.shape[0] != runtimes.shape[0]:
        raise ValueError("Sample X and runtimes must be the same length")

    return train_gaussian_process(sample_X, runtimes, X_bounds)


#def train_failure_model(sample_X, failures, X_bounds):
#    if sample_X.shape[0] != failures.shape[0]:
#        raise ValueError("Sample X and runtimes must be the same length")
#
#    return train_gaussian_process(sample_X, runtimes, X_bounds)


def train_gaussian_process(
    sample_X, sample_y, X_bounds, current_X=None, nu=1.5, max_samples=100
):
    """
    Trains a Gaussian Process function from sample_X, sample_y data

    Handles the case where there are other training runs in flight (current_X)

        Arguments:
            sample_X - vector of already evaluated sets of hyperparameters
            sample_y - vector of already evaluated loss function values
            X_bounds - minimum and maximum values for every dimension of X
            current_X - hyperparameters currently being explored
            nu - input to the Matern function, higher numbers make it smoother 0.5, 1.5, 2.5 are good values
             see http://scikit-learn.org/stable/modules/generated/sklearn.gaussian_process.kernels.Matern.html

        Returns:
            gp - the gaussian process function
            y_mean - mean
            y_stddev - stddev

            To make a prediction with gp on real world data X, need to call:
            (gp.predict(X) * y_stddev) + y_mean

    """
    if current_X is not None:
        current_X = np.array(current_X)
        if len(current_X.shape) != 2:
            raise ValueError("Current X must be a 2 dimensional array")

        # we can't let the current samples be bigger than max samples
        # because we need to use some real samples to build the curve
        if current_X.shape[0] > max_samples - 5:
            print(
                "current_X is bigger than max samples - 5 so dropping some currently running parameters"
            )
            current_X = current_X[:(max_samples - 5), :]
    if len(sample_y.shape) != 1:
        raise ValueError("Sample y must be a 1 dimensional array")

    if sample_X.shape[0] != sample_y.shape[0]:
        raise ValueError(
            "Sample X and sample y must be the same size {} {}".format(
                sample_X.shape[0], sample_y.shape[0]
            )
        )

    if X_bounds is not None and sample_X.shape[1] != len(X_bounds):
        raise ValueError(
            "Bounds must be the same length as Sample X's second dimension"
        )

    # gaussian process takes a long time to train, so if there's more than max_samples
    # we need to sample from it
    if sample_X.shape[0] > max_samples:
        sample_indices = np.random.randint(sample_X.shape[0], size=max_samples)
        X = sample_X[sample_indices]
        y = sample_y[sample_indices]
    else:
        X = sample_X
        y = sample_y
    gp, y_mean, y_stddev = fit_normalized_gaussian_process(X, y, nu=nu)
    if current_X is not None:
        # if we have some hyperparameters running, we pretend that they return
        # the prediction of the function we've fit
        X = np.append(X, current_X, axis=0)
        current_y_fantasy = (gp.predict(current_X) * y_stddev) + y_mean
        y = np.append(y, current_y_fantasy)
        gp, y_mean, y_stddev = fit_normalized_gaussian_process(X, y, nu=nu)
    return gp, y_mean, y_stddev


def filter_weird_values(sample_X, sample_y):
    is_row_finite = ~(np.isnan(sample_X).any(axis=1) | np.isnan(sample_y))
    sample_X = sample_X[is_row_finite, :]
    sample_y = sample_y[is_row_finite]
    return sample_X, sample_y


def next_sample(
    sample_X,
    sample_y,
    X_bounds=None,
    runtimes=None,
    failures=None,
    current_X=None,
    nu=1.5,
    max_samples_for_gp=100,
    improvement=0.01,
    num_points_to_try=1000,
    opt_func="expected_improvement",
    test_X=None,
):
    """
        Calculates the best next sample to look at via bayesian optimization.

        Check out https://arxiv.org/pdf/1206.2944.pdf
         for explanation of bayesian optimization

        Arguments:
            sample_X - 2d array of already evaluated sets of hyperparameters
            sample_y - 1d array of already evaluated loss function values
            X_bounds - 2d array minimum and maximum values for every dimension of X

            runtimes - vector of length sample_y - should be the time taken to train each model in sample X
            failures - vector of length sample_y - should be True for models where training failed and False where
                training succeeded.  This model will throw out NaNs and Infs so if you want it to avaoid
                failure values for X, use this failure vector.

            current_X - hyperparameters currently being explored
            nu - input to the Matern function, higher numbers make it smoother 0.5, 1.5, 2.5 are good values
             see http://scikit-learn.org/stable/modules/generated/sklearn.gaussian_process.kernels.Matern.html

            max_samples_for_gp - maximum samples to consider (since algo is O(n^3)) for performance, but also adds some randomness
            improvement - amount of improvement to optimize for -- higher means take more exploratory risks
            num_points_to_try - number of X values to try when looking for value with highest
                        expected probability of improvement
            opt_func - one of {"expected_improvement", "prob_of_improvement"} - whether to optimize expected
                improvement of probability of improvement.  Expected improvement is generally better - may want
                to remove probability of improvement at some point.  (But I think prboability of improvement
                is a little easier to calculate)
            test_X - X values to test when looking for the best values to try

        Returns:
            suggested_X - X vector to try running next
            suggested_X_prob_of_improvement - probability of the X vector beating the current best
            suggested_X_predicted_y - predicted output of the X vector
            test_X - 2d array of length num_points_to_try by num features: tested X values
            y_pred - 1d array of length num_points_to_try: predicted values for test_X
            y_pred_std - 1d array of length num_points_to_try: predicted std deviation for test_X
            prob_of_improve 1d array of lenth num_points_to_try: predicted porbability of improvement
            prob_of_failure 1d array of predicted probabilites of failure
            expected_runtime 1d array of expected runtimes
    """
    # Sanity check the data
    sample_X = np.array(sample_X)
    sample_y = np.array(sample_y)
    if test_X is not None:
        test_X = np.array(test_X)
    if len(sample_X.shape) != 2:
        raise ValueError("Sample X must be a 2 dimensional array")

    if len(sample_y.shape) != 1:
        raise ValueError("Sample y must be a 1 dimensional array")

    if sample_X.shape[0] != sample_y.shape[0]:
        raise ValueError("Sample X and y must be same length")

    if test_X is not None:
        # if test_X is set, usually this is for simulation/testing
        if X_bounds is not None:
            raise ValueError("Can't set test_X and X_bounds")

    else:
        # normal case where we randomly sample our test_X
        if X_bounds is None:
            raise ValueError("Must pass in test_X or X_bounds")

    filtered_X, filtered_y = filter_weird_values(sample_X, sample_y)
    # We train our runtime prediction model on *filtered_X* throwing out the sample points with
    # NaN values because they might break our runtime predictor
    runtime_model = None
    if runtimes is not None:
        runtime_filtered_X, runtime_filtered_runtimes = filter_weird_values(
            sample_X, runtimes
        )
        if runtime_filtered_X.shape[0] >= 2:
            runtime_model, runtime_model_mean, runtime_model_stddev = train_runtime_model(
                runtime_filtered_X, runtime_filtered_runtimes
            )
    # We train our failure model on *sample_X*, all the data including NaNs
    # This is *different* than the runtime model.
    failure_model = None
    if failures is not None and sample_X.shape[0] >= 2:
        failure_filtered_X, failure_filtered_runtimes = filter_weird_values(
            sample_X, failures
        )
        if failure_filtered_X.shape[0] >= 2:
            failure_model, failure_model_mean, failure_model_stddev = train_runtime_model(
                failure_filtered_X, failure_filtered_runtimes
            )
    # we can't run this algothim with less than two sample points, so we'll
    # just return a random point
    if filtered_X.shape[0] < 2:
        if test_X is not None:
            # pick a random row from test_X
            row = np.random.choice(test_X.shape[0])
            X = test_X[row, :]
        else:
            X = random_sample(X_bounds, 1)[0]
        if filtered_X.shape[0] < 1:
            prediction = 0.0
        else:
            prediction = filtered_y[0]
        return X, 1.0, prediction, None, None, None, None, None, None

    # build the acquisition function
    gp, y_mean, y_stddev, = train_gaussian_process(
        filtered_X, filtered_y, X_bounds, current_X, nu, max_samples_for_gp
    )
    num_test_samples = 1000
    # Look for the minimum value of our fitted-target-function + (kappa * fitted-target-std_dev)
    if test_X is None:  # this is the usual case
        test_X = random_sample(X_bounds, num_test_samples)
    y_pred, y_pred_std = gp.predict(test_X, return_std=True)
    if failure_model is None:
        prob_of_failure = [0.0] * len(test_X)
    else:
        prob_of_failure = failure_model.predict(
            test_X
        ) * failure_model_stddev + failure_model_mean
    if runtime_model is None:
        expected_runtime = [0.0] * len(test_X)
    else:
        expected_runtime = runtime_model.predict(
            test_X
        ) * runtime_model_stddev + runtime_model_mean
    # best value of y we've seen so far.  i.e. y*
    min_unnorm_y = np.min(filtered_y)
    # hack for dealing with predicted std of 0
    epsilon = 0.00000001
    if opt_func == "probability_of_improvement":
        # might remove the norm_improvement at some point
        # find best chance of an improvement by "at least norm improvement"
        # so if norm_improvement is zero, we are looking for best chance of any
        # improvment over the best result observerd so far.
        #norm_improvement = improvement / y_stddev
        min_norm_y = (min_unnorm_y - y_mean) / y_stddev - improvement
        distance = (y_pred - min_norm_y)
        std_dev_distance = (y_pred - min_norm_y) / (y_pred_std + epsilon)
        prob_of_improve = sigmoid(-std_dev_distance)
        best_test_X_index = np.argmax(prob_of_improve)
    elif opt_func == "expected_improvement":
        min_norm_y = (min_unnorm_y - y_mean) / y_stddev
        Z = -(y_pred - min_norm_y) / (y_pred_std + epsilon)
        prob_of_improve = scipy_stats.norm.cdf(Z)
        e_i = -(y_pred - min_norm_y) * scipy_stats.norm.cdf(Z) + y_pred_std * scipy_stats.norm.pdf(
            Z
        )
        best_test_X_index = np.argmax(e_i)
    # TODO: support expected improvement per time by dividing e_i by runtime
    suggested_X = test_X[best_test_X_index]
    suggested_X_prob_of_improvement = prob_of_improve[best_test_X_index]
    suggested_X_predicted_y = y_pred[best_test_X_index] * y_stddev + y_mean
    unnorm_y_pred = y_pred * y_stddev + y_mean
    unnorm_y_pred_std = y_pred_std * y_stddev
    return (
        suggested_X,
        suggested_X_prob_of_improvement,
        suggested_X_predicted_y,
        test_X,
        unnorm_y_pred,
        unnorm_y_pred_std,
        prob_of_improve,
        prob_of_failure,
        expected_runtime,
    )


def target(x):
    return np.exp(-(x - 2) ** 2) + np.exp(-(x - 6) ** 2 / 10) + 1 / (x ** 2 + 1)


class BayesianSearch(Search):
    def __init__(self, minimum_improvement=0.1):
        self.minimum_improvement = minimum_improvement

    def next_run(self, sweep):
        if 'parameters' not in sweep['config']:
            raise ValueError('Bayesian search requires "parameters" section')
        if 'metric' not in sweep['config']:
            raise ValueError('Bayesian search requires "metric" section')

        config = sweep['config']['parameters']
        params = HyperParameterSet.from_config(config)

        sample_X = []
        sample_y = []
        current_X = []
        y = []

        params.index_searchable_params()

        # X_bounds = [[0., 1.]] * len(self.searchable_params)
        # params.numeric_bounds()
        X_bounds = [[0., 1.]] * len(params.searchable_params)

        runs = sweep['runs']

        # we calc the max metric to put as the metric for failed runs
        # so that our bayesian search stays away from them
        max_metric = 0.
        if any(run.state == "finished" for run in runs):
            # for run in runs:
            #    print("DEBUG0", run)
            max_metric = max([self._metric_from_run(sweep['config'], run, default=0.) for run in runs
                              if run.state == "finished"])

        X_norms = params.convert_runs_to_normalized_vector(runs)
        for i in range(len(runs)):
            run = runs[i]
            X_norm = X_norms[i]
            if run.state == "finished":
                # run is complete
                #print("DEBUG0.1", run)
                metric = self._metric_from_run(sweep['config'], run, default=max_metric)
                if math.isnan(metric):
                    metric = max_metric
                y.append(metric)
                sample_X.append(X_norm)
            elif run.state == "running":
                # run is in progress
                # we wont use the metric, but we should pass it into our optimizer to
                # account for the fact that it is running
                current_X.append(X_norm)
            elif run.state == "failed" or run.state == "crashed" or run.state == "killed":
                # run failed, but we're still going to use it
                # maybe we should be smarter about this
                y.append(max_metric)
                sample_X.append(X_norm)
            else:
                raise ValueError("Run is in unknown state")

        if len(sample_X) == 0:
            sample_X = np.empty([0, 0])

        if len(current_X) == 0:
            current_X = None
        else:
            np.array(current_X)
        (try_params, success_prob, pred,
            test_X, y_pred, y_pred_std, prob_of_improve,
            prob_of_failure, expected_runtime) = next_sample(
                np.array(sample_X),
                np.array(y), X_bounds,
                current_X=current_X, improvement=self.minimum_improvement)

        # convert the parameters from vector of [0,1] values
        # to the original ranges

        for param in params:
            if param.type == HyperParameter.CONSTANT:
                continue

            # try_value = try_params[params.param_names_to_index[param.name]]
            # if param.type == HyperParameter.CATEGORICAL:
            #     param.value = param.values[int(try_value)]
            # elif param.type == HyperParameter.INT_UNIFORM:
            #     param.value = int(try_value)
            # elif param.type == HyperParameter.UNIFORM:
            #     param.value = try_value
            try_value = try_params[params.param_names_to_index[param.name]]
            param.value = param.ppf(try_value)

        metric_name = sweep['config']['metric']['name']

        ret_dict = params.to_config()
        info = {}
        info['predictions'] = {metric_name: pred}
        info['success_probability'] = success_prob
        if test_X is not None:
            info['acq_func'] = {}
            info['acq_func']['sample_x'] = params.denormalize_vector(test_X)
            info['acq_func']['y_pred'] = y_pred
            info['acq_func']['y_pred_std'] = y_pred_std
            info['acq_func']['score'] = prob_of_improve

        return ret_dict, info



"""
monkeypatch: patch code to add tensorboard hooks
"""

import os
import socket
import re

import wandb


TENSORBOARD_C_MODULE = "tensorflow.python.ops.gen_summary_ops"
TENSORBOARD_X_MODULE = "tensorboardX.writer"
TENSORFLOW_PY_MODULE = "tensorflow.python.summary.writer.writer"
TENSORBOARD_WRITER_MODULE = "tensorboard.summary.writer.event_file_writer"
TENSORBOARD_PYTORCH_MODULE = "torch.utils.tensorboard.writer"


def unpatch():
    for module, method in wandb.patched["tensorboard"]:
        writer = wandb.util.get_module(module)
        setattr(writer, method, getattr(writer, "orig_{}".format(method)))
    wandb.patched["tensorboard"] = []


def patch(save=None, tensorboardX=None, pytorch=None, root_logdir=None):
    if len(wandb.patched["tensorboard"]) > 0:
        raise ValueError(
            "Tensorboard already patched, remove sync_tensorboard=True from wandb.init or only call wandb.tensorboard.patch once."
        )

    # TODO: Some older versions of tensorflow don't require tensorboard to be present.
    # we may want to lift this requirement, but it's safer to have it for now
    wandb.util.get_module("tensorboard", required="Please install tensorboard package")
    c_writer = wandb.util.get_module(TENSORBOARD_C_MODULE)
    py_writer = wandb.util.get_module(TENSORFLOW_PY_MODULE)
    tb_writer = wandb.util.get_module(TENSORBOARD_WRITER_MODULE)
    pt_writer = wandb.util.get_module(TENSORBOARD_PYTORCH_MODULE)
    tbx_writer = wandb.util.get_module(TENSORBOARD_X_MODULE)

    if not pytorch and not tensorboardX and c_writer:
        _patch_tensorflow2(
            writer=c_writer,
            module=TENSORBOARD_C_MODULE,
            save=save,
            root_logdir=root_logdir,
        )
    # This is for tensorflow <= 1.15 (tf.compat.v1.summary.FileWriter)
    if py_writer:
        _patch_file_writer(
            writer=py_writer,
            module=TENSORFLOW_PY_MODULE,
            save=save,
            root_logdir=root_logdir,
        )
    if tb_writer:
        _patch_file_writer(
            writer=tb_writer,
            module=TENSORBOARD_WRITER_MODULE,
            save=save,
            root_logdir=root_logdir,
        )
    if pt_writer:
        _patch_file_writer(
            writer=pt_writer,
            module=TENSORBOARD_PYTORCH_MODULE,
            save=save,
            root_logdir=root_logdir,
        )
    if tbx_writer:
        _patch_file_writer(
            writer=tbx_writer,
            module=TENSORBOARD_X_MODULE,
            save=save,
            root_logdir=root_logdir,
        )
    if not c_writer and not tb_writer and not tb_writer:
        wandb.termerror("Unsupported tensorboard configuration")


def _patch_tensorflow2(
    writer, module, save=None, root_logdir=None,
):
    # This configures TensorFlow 2 style Tensorboard logging
    old_csfw_func = writer.create_summary_file_writer
    logdir_hist = []

    def new_csfw_func(*args, **kwargs):
        logdir = (
            kwargs["logdir"].numpy().decode("utf8")
            if hasattr(kwargs["logdir"], "numpy")
            else kwargs["logdir"]
        )
        logdir_hist.append(logdir)
        root_logdir_arg = root_logdir

        if len(set(logdir_hist)) > 1 and root_logdir is None:
            wandb.termwarn(
                'When using several event log directories, please call wandb.tensorboard.patch(root_logdir="...") before wandb.init'
            )
        # if the logdir containts the hostname, the writer was not given a logdir. In this case, the generated logdir
        # is genetered and ends with the hostname, update the root_logdir to match.
        hostname = socket.gethostname()
        search = re.search(r"-\d+_{}".format(hostname), logdir)
        if search:
            root_logdir_arg = logdir[: search.span()[1]]
        elif root_logdir is not None and not os.path.abspath(logdir).startswith(
            os.path.abspath(root_logdir)
        ):
            wandb.termwarn(
                "Found logdirectory outside of given root_logdir, dropping given root_logdir for eventfile in {}".format(
                    logdir
                )
            )
            root_logdir_arg = None

        _notify_tensorboard_logdir(logdir, save=save, root_logdir=root_logdir_arg)
        return old_csfw_func(*args, **kwargs)

    writer.orig_create_summary_file_writer = old_csfw_func
    writer.create_summary_file_writer = new_csfw_func
    wandb.patched["tensorboard"].append([module, "create_summary_file_writer"])


def _patch_file_writer(writer, module, save=None, root_logdir=None):
    # This configures non-TensorFlow Tensorboard logging, or tensorflow <= 1.15
    old_efw_class = writer.EventFileWriter

    logdir_hist = []

    class TBXEventFileWriter(old_efw_class):
        def __init__(self, logdir, *args, **kwargs):
            logdir_hist.append(logdir)
            root_logdir_arg = root_logdir
            if len(set(logdir_hist)) > 1 and root_logdir is None:
                wandb.termwarn(
                    'When using several event log directories, please call wandb.tensorboard.patch(root_logdir="...") before wandb.init'
                )

            # if the logdir containts the hostname, the writer was not given a logdir. In this case, the generated logdir
            # is genetered and ends with the hostname, update the root_logdir to match.
            hostname = socket.gethostname()
            search = re.search(r"-\d+_{}".format(hostname), logdir)
            if search:
                root_logdir_arg = logdir[: search.span()[1]]

            elif root_logdir is not None and not os.path.abspath(logdir).startswith(
                os.path.abspath(root_logdir)
            ):
                wandb.termwarn(
                    "Found logdirectory outside of given root_logdir, dropping given root_logdir for eventfile in {}".format(
                        logdir
                    )
                )
                root_logdir_arg = None

            _notify_tensorboard_logdir(logdir, save=save, root_logdir=root_logdir_arg)

            super(TBXEventFileWriter, self).__init__(logdir, *args, **kwargs)

    writer.orig_EventFileWriter = old_efw_class
    writer.EventFileWriter = TBXEventFileWriter
    wandb.patched["tensorboard"].append([module, "EventFileWriter"])


def _notify_tensorboard_logdir(logdir, save=None, root_logdir=None):
    wandb.run._tensorboard_callback(logdir, save=save, root_logdir=root_logdir)
